from __future__ import annotations

from typing import Any

from sqlglot import expressions as exp
from mysql_mimic.variables import DEFAULT


# Mapping of transaction characteristic from SET TRANSACTION statements to their corresponding system variable
TRANSACTION_CHARACTERISTICS = {
    "ISOLATION LEVEL REPEATABLE READ": ("transaction_isolation", "REPEATABLE-READ"),
    "ISOLATION LEVEL READ COMMITTED": ("transaction_isolation", "READ-COMMITTED"),
    "ISOLATION LEVEL READ UNCOMMITTED": ("transaction_isolation", "READ-UNCOMMITTED"),
    "ISOLATION LEVEL SERIALIZABLE": ("transaction_isolation", "SERIALIZABLE"),
    "READ WRITE": ("transaction_read_only", False),
    "READ ONLY": ("transaction_read_only", True),
}


def setitem_kind(setitem: exp.SetItem) -> str:
    kind = setitem.text("kind")
    if not kind:
        return "VARIABLE"

    if kind in {"GLOBAL", "PERSIST", "PERSIST_ONLY", "SESSION", "LOCAL"}:
        return "VARIABLE"

    return kind


def value_to_expression(value: Any) -> exp.Expression:
    if value is True:
        return exp.true()
    if value is False:
        return exp.false()
    if value is None:
        return exp.null()
    if isinstance(value, (int, float)):
        return exp.Literal.number(value)
    return exp.Literal.string(str(value))


def expression_to_value(expression: exp.Expression) -> Any:
    if expression == exp.true():
        return True
    if expression == exp.false():
        return False
    if expression == exp.null():
        return None
    if isinstance(expression, exp.Literal) and not expression.args.get("is_string"):
        try:
            return int(expression.this)
        except ValueError:
            return float(expression.this)
    if isinstance(expression, exp.Literal):
        return expression.name
    if expression.name == "DEFAULT":
        return DEFAULT
    if expression.name == "ON":
        return True
    if expression.name == "OFF":
        return False
    return expression.name



from datetime import timezone, timedelta

import pytest

from mysql_mimic.errors import MysqlError
from mysql_mimic.variables import parse_timezone


def test_parse_timezone() -> None:
    assert timezone(timedelta()) == parse_timezone("UTC")
    assert timezone(timedelta()) == parse_timezone("+00:00")
    assert timezone(timedelta(hours=1)) == parse_timezone("+01:00")
    assert timezone(timedelta(hours=-1)) == parse_timezone("-01:00")

    # Implicitly test cache
    assert timezone(timedelta(hours=-1)) == parse_timezone("-01:00")

    with pytest.raises(MysqlError):
        parse_timezone("whoops")



from __future__ import annotations

import inspect
import sys
from collections.abc import Iterator
import random
from typing import List, TypeVar, AsyncIterable, Iterable, AsyncIterator, cast
import string

from sqlglot import expressions as exp
from sqlglot.optimizer.scope import traverse_scope


T = TypeVar("T")

# MySQL Connector/J uses ASCII to decode nonce
SAFE_NONCE_CHARS = (string.ascii_letters + string.digits).encode()


class seq(Iterator):
    """Auto-incrementing sequence with an optional maximum size"""

    def __init__(self, size: int | None = None):
        self.size = size
        self.value = 0

    def __next__(self) -> int:
        value = self.value
        self.value = self.value + 1
        if self.size:
            self.value = self.value % self.size
        return value

    def reset(self) -> None:
        self.value = 0


def xor(a: bytes, b: bytes) -> bytes:
    # Fast XOR implementation, according to https://stackoverflow.com/questions/29408173/byte-operations-xor-in-python
    a, b = a[: len(b)], b[: len(a)]
    int_b = int.from_bytes(b, sys.byteorder)
    int_a = int.from_bytes(a, sys.byteorder)
    int_enc = int_b ^ int_a
    return int_enc.to_bytes(len(b), sys.byteorder)


def nonce(nbytes: int) -> bytes:
    return bytes(
        bytearray(
            [random.SystemRandom().choice(SAFE_NONCE_CHARS) for _ in range(nbytes)]
        )
    )


def find_tables(expression: exp.Expression) -> List[exp.Table]:
    """Find all tables in an expression"""
    if isinstance(expression, (exp.Subqueryable, exp.Subquery)):
        return [
            source
            for scope in traverse_scope(expression)
            for source in scope.sources.values()
            if isinstance(source, exp.Table)
        ]
    return []


def find_dbs(expression: exp.Expression) -> List[str]:
    """Find all database names in an expression"""
    return [table.text("db") for table in find_tables(expression)]


def dict_depth(d: dict) -> int:
    """
    Get the nesting depth of a dictionary.
    For example:
        >>> dict_depth(None)
        0
        >>> dict_depth({})
        1
        >>> dict_depth({"a": "b"})
        1
        >>> dict_depth({"a": {}})
        2
        >>> dict_depth({"a": {"b": {}}})
        3
    Args:
        d (dict): dictionary
    Returns:
        int: depth
    """
    try:
        return 1 + dict_depth(next(iter(d.values())))
    except AttributeError:
        # d doesn't have attribute "values"
        return 0
    except StopIteration:
        # d.values() returns an empty sequence
        return 1


async def aiterate(iterable: AsyncIterable[T] | Iterable[T]) -> AsyncIterator[T]:
    """Iterate either an async iterable or a regular iterable"""
    if inspect.isasyncgen(iterable):
        async for item in iterable:
            yield item
    else:
        for item in cast(Iterable, iterable):
            yield item



from typing import Any

import pytest

from mysql_mimic import ColumnType
from mysql_mimic.errors import MysqlError
from mysql_mimic.results import ensure_result_set


async def gen_rows() -> Any:
    yield 1, None, None
    yield None, "2", None
    yield None, None, None


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "result, column_types",
    [
        (
            (gen_rows(), ["a", "b", "c"]),
            [ColumnType.LONGLONG, ColumnType.STRING, ColumnType.NULL],
        ),
    ],
)
async def test_ensure_result_set_columns(result: Any, column_types: Any) -> None:
    result_set = await ensure_result_set(result)
    assert [c.type for c in result_set.columns] == column_types


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "result",
    [
        [1, 2],
        ([[1, 2]], ["a", "b"], ["a", "b"]),
    ],
)
async def test_ensure_result_set__invalid(result: Any) -> None:
    with pytest.raises(MysqlError):
        await ensure_result_set(result)
